#include "FCStructure.H"
#include "Blackboard.H"

/*
	Clears the structure. Should be called before a new coordination begins.
*/
void FCStructure::clear() {
	items.clear();
}

/*
	Returns true if at least one possible fire combination is added to the structure.
*/
bool FCStructure::hasEntries() {
	if (items.size() > 0) {
		return true;
	}

	return false;
}

/*
	Adds a possible attack combination. The unit [unit] can attack [target].
*/
void FCStructure::addAttackPossibility(uint4 unitId, GameObjPtr unit, uint4 targetId, GameObjPtr target) {
	bool found = false;

	//Step 1 - See if the target unit already has been added to the structure
	vector<FCStructureItem>::iterator it = items.begin();
	while (it != items.end()) {
		if ((*it).targetId == targetId) {
			//Target exists in other attack possibilities. Update the element with the current possibility.
			(*it).addAttacker(unitId, unit);
			found = true;
		}
		it++;
	}

	//Step 2 - Add new target unit to the structure (if it isn't already present)
	if (!found) {
		int HPleft = target->get_int("hp");
		FCStructureItem mItem;
		mItem.targetId = targetId;
		mItem.targetObj = target;
		mItem.targetUnitType = Blackboard::getType(target);
		mItem.HPleft = HPleft;
		mItem.addAttacker(unitId, unit);
		items.push_back(mItem);
	}
}

/*
	After all attack possibilities has been added to the structure this method shall
	be called to do the actual fire coordination.
	After processing each friendly unit has 0 or 1 targets to fire at.
*/
void FCStructure::findAttackScheme() {
	if (!hasEntries()) return;

	sort(items.begin(), items.end());
	
	//Step 1 - Check if it's possible to completely destroy enemy units.
	//This attack possibilities are of highest priority.
	for (uint4 i = 0; i < items.size(); i++) {
		int destroy = items[i].canBeDestroyed();
		if (destroy >= 0) {
			if (destroy > 0) {
				//There are more units attacking this target than necessary. Remove.
				removeUnnecessary(&items[i], destroy);
			}
			removeAttackers(&items[i]);
		}
	}

	//Step 2 - Attack weakest targets or (if HPleft are equal) targets many units can fire at.
	for (uint4 i = 0; i < items.size(); i++) {
		if (items[i].canBeDestroyed() == -1) {
			//Parse the rest of the units (i.e. those that cannot be destroyed)
			removeAttackers(&items[i]);
		}
	}

	//Step 3 - Remove units that cannot be attacked after previous steps
	removeEmptyEntries();
}

/*
	If there are more attacking units than what is needed to destroy the
	target this method removes the unnecessary attacking units.

	To decide which entries to remove the method calculates the number of occurences
	the attacker unit has in the whole data structure, and the ones with max
	occurences (which are most likely to be in other combinations) are removed.
*/
void FCStructure::removeUnnecessary(FCStructureItemPtr item, int noToRemove) {
	for (int n = 0; n < noToRemove; n++) {
		int max = -1;
		uint4 maxId = 0;

		//Find the attacking unit with most occurences in the data structure
		for (uint4 i = 0; i < item->getNoAttackers(); i++) {
			uint4 unitId = item->getAttackerId(i);
			int occurences = calculateOccurences(unitId);
			if (max == -1 || (occurences > max)) {
				max = occurences;
				maxId = unitId;
			}
		}

		//Remove that unit from attacking this target
		item->removeAttacker(maxId);
	}
}

/*
	Calculates the number of occurences an attacking unit has in the whole
	data structure.
*/
int FCStructure::calculateOccurences(uint4 unitId) {
	int occurences = 0;
	for (uint4 i = 0; i < items.size(); i++) {
		if(items[i].attackedBy(unitId)) {
			occurences++;
		}
	}

	return occurences;
}

/*
	Returns the object Id of the target for the specified unit. Returns -1 if no targets are available.
*/
int FCStructure::getTargetId(uint4 unitId) {
	for (uint4 i = 0; i < items.size(); i++) {
		if(items[i].attackedBy(unitId)) {
			return items[i].targetId;
		}

	}

	return -1;
}

/*
	Returns the GameObj of the target for the specified unit. Returns null if no targets are available.
*/
GameObj* FCStructure::getTargetObj(uint4 unitId) {
	for (uint4 i = 0; i < items.size(); i++) {
		if(items[i].attackedBy(unitId)) {
			return items[i].targetObj;
		}

	}

	return NULL;
}

/*
	Removes targets that after coordination processing aren't attacked by any units.
*/
void FCStructure::removeEmptyEntries() {
	for (uint4 i = 0; i < items.size(); i++) {
		if (items[i].getNoAttackers() == 0) {
			items.erase(items.begin() + i);
			i--;
		}
	}
}

/*
	This method is used during coordination processing.
	The logic has decided that the specified unit/target combination (FCStructureItem) shall remain after
	the coordination is finished and this method removes the attacking units from other unit/target combinations.
*/
void FCStructure::removeAttackers(FCStructureItemPtr item) {
	for (uint4 i = 0; i < item->getNoAttackers(); i++) {
		uint4 attackerId = item->getAttackerId(i);
		removeAttacker(item->targetId, attackerId);
	}
}

/*
	Removes the unit [unitId] from attacking other targets than [currentTargetId].
*/
void FCStructure::removeAttacker(uint4 currentTargetId, uint4 unitId) {
	vector<FCStructureItem>::iterator it = items.begin();
	while (it != items.end()) {
		if ((*it).targetId != currentTargetId) {
			//Remove attacker if the target is different than current target
			(*it).removeAttacker(unitId);
		}
		it++;
	}
}

/*
	Dumps the data structure to the console window for debugging purpose.
*/
void FCStructure::dump() {
	cout << "DUMP" << endl;

	vector<FCStructureItem>::iterator it = items.begin();
	while (it != items.end()) {
		(*it).dump();
		it++;
	}

	cout << "" << endl;
}
