#ifndef PF_H
#define PF_H

#include "Trail.H"
#include "GameObj.H"

using namespace std;

struct ObjPos {
	uint4 objId;
	sint4 type;

	int x;
	int y;
	int HP;
	int gametick;
	bool destroyed;
	bool visible;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	void update(uint4 mGametick, int mX, int mY) {
		destroyed = false;
		x = mX;
		y = mY;
		gametick = mGametick;
	}

	bool isMarine() {
		if (type == 1) return true;
		return false;
	}

	bool isTank() {
		if (type == 0) return true;
		return false;
	}

	real8 getHealthPct() {
		if (isTank()) {
			return (real8)HP / (real8)150;
		}
		if (isMarine()) {
			return (real8)HP / (real8)80;
		}
		return 0;
	}
};

struct BasePos {
	uint4 objId;
	
	int x;
	int y;
	int x1;
	int x2;
	int y1;
	int y2;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	bool isPositionInside(int cX, int cY) {
		if (cX >= x1 - 7 && cX <= x2 + 7) {
			if (cY >= y1 - 7 && cY <= y2 + 7) {
				return true;
			}
		}
		return false;
	}
};

struct ObjFuturePos {
	uint4 objId;
	sint4 type;
	
	int x;
	int y;
	int nextX;
	int nextY;
	int HP;
	int gametick;
	bool destroyed;

	Trail trail;

	real8 prob;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	void update(uint4 mGametick, int mX, int mY) {
		destroyed = false;
		x = mX;
		y = mY;
		nextX = -1;
		nextY = -1;
		gametick = mGametick;
	}

	bool hasNextPos() {
		if (nextX == -1) return false;
		if (nextY == -1) return false;
		return true;
	}

	bool isMarine() {
		if (type == 1) return true;
		return false;
	}

	bool isTank() {
		if (type == 0) return true;
		return false;
	}

	real8 getHealthPct() {
		if (isTank()) {
			return (real8)HP / (real8)150;
		}
		if (isMarine()) {
			return (real8)HP / (real8)80;
		}
		return 0;
	}
};

struct ObjHistoryPos {
	uint4 objId;

	int x_minusOne;
	int y_minusOne;
	int gametick_minusOne;

	int x_minusTwo;
	int y_minusTwo;
	int gametick_minusTwo;

	bool destroyed;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	void update(uint4 gametick, int x, int y) {
		destroyed = false;

		if (x != x_minusOne || y != y_minusOne) {
			x_minusTwo = x_minusOne;
			x_minusOne = x;
			
			y_minusTwo = y_minusOne;
			y_minusOne = y;

			gametick_minusTwo = gametick_minusOne;
			gametick_minusOne = gametick;
		}
	}
};

struct Position {
	int x;
	int y;

	bool found() {
		if (x == -1) return false;
		if (y == -1) return false;
		return true;
	}
};

/*
	This class is responsible for calculating the correct potential in a point at runtime. It is used in Addition 4.

	To speed up iteration of game objects the class stores each object in a vector instead of iterating through the large
	GameObj lists returned from the game server.
	Each game object is stored in one of five vectors:
	sheep - Neutral units (i.e. sheep).
	enemies - Enemy tanks or marines.
	units - Own tanks or marines.
	bases - Own bases.
	enemyBases - Enemy bases.

	There are four types of position objects (with different functionality) to use:
	ObjPos - The basic object. Contains position, gametick and a destroyed flag.
	BasePos - Same as ObjPos but contains an area instead of a single position.
	ObjHistoryPos - Contains the last two positions for an unit. Can be used to predict where the unit might move next.
	ObjFuturePos - Contains position as well as the position a unit is moving to in next game frame. Can be used to avoid unit collision.
*/
class PF {
	
	typedef GameObj* GameObjPtr;
	
	public:

		PF();

		void updateProbabilities();

		real8 getProbability(uint4 objId);

		int getOU();
		int getEU();
		int getOB();
		int getEB();

		/*
			Sets all own and enemy units as destroyed. Shall be called before updating the unit positions
			each frame. A unit which has its position updated are set as not destroyed, and after the
			position updates all units which still has the destroyed flag set can be removed.
		*/
		void setAllDestroyed();

		/*
			Checks which enemy units visible by own units.
		*/
		void checkVisibility();
		
		/*
			Removes destroyed own and enemy units.
			Must be used with the setAllDestroyed() method.
		*/
		void cleanUp();

		/*
			Updates the position for sheep [objId].
		*/
		void updateSheepPosition(uint4 game_tick, uint4 objId, int x, int y);

		/*
			Updates the position for enemy unit [objId].
		*/
		void updateEnemyUnitPosition(uint4 game_tick, uint4 objId, sint4 unitType, int x, int y, int cHP);

		/*
			Updates the position for own unit [objId].
		*/
		void updateUnitPosition(uint4 game_tick, uint4 objId, sint4 unitType, int x, int y, int cHP);

		/*
			Sets the position the unit [objId] is moving to.
		*/
		void setNextUnitPosition(uint4 objId, int nextX, int nextY);
		
		/*
			Returns the potential generated by sheep in the point (pX,pY).
		*/
		real8 getSheepCharge(int pX, int pY, uint4 game_tick);

		/*
			Returns the potential generated by enemy units in the point (pX,pY).
		*/
		real8 getEnemyUnitCharge(int pX, int pY, bool offensive, uint4 game_tick);

		/*
			Returns the potential generated by own units in the point (pX,pY).
		*/
		real8 getUnitCharge(int pX, int pY, uint4 objId, uint4 game_tick);

		/*
			Returns the potential generated by the marine coordination field in the point (pX,pY).
		*/
		real8 getMarineCoordCharge(int pX, int pY, uint4 game_tick, int map_width);

		/*
			Removes all elements in the own- and enemy bases vectors.
		*/
		void clearBaseLists();

		/*
			Adds the base [objId] to a bases vector.
		*/
		void addBase(uint4 objId, int x, int y, int x1, int y1, int x2, int y2, bool own);

		/*
			Returns the potential generated by own bases in the point (pX,pY).
		*/
		real8 getBaseCharge(int pX, int pY);

		/*
			Returns the potential generated by enemy bases in the point (pX,pY).
		*/
		real8 getEnemyBaseCharge(int pX, int pY);

		vector<ObjFuturePos>* getUnitsList();
		vector<ObjPos>* getEnemyUnitsList();
		vector<BasePos>* getEnemyBasesList();

		//Vectors holding the game objects.
		vector<ObjPos> sheep;
		vector<ObjPos> enemies;
		vector<ObjFuturePos> units;
		vector<BasePos> bases;
		vector<BasePos> enemyBases;

	private:
	
		/*
			Returns the potential a sheep at (objX,objY) generates in the point (x,y).
		*/
		real8 calcSheepCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential a marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcOffensiveMarineCharge(int x, int y, int objX, int objY);
		
		/*
			Returns the potential a marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcDefensiveMarineCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential a tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcOffensiveTankCharge(int x, int y, int objX, int objY);
		
		/*
			Returns the potential a tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcDefensiveTankCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential an own marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcMarineUnitCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential an own tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcTankUnitCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential an own base at (objX,objY) generates in the point (x,y).
		*/
		real8 calcBaseCharge(int x, int y, int objX, int objY, bool inside);

		/*
			Returns the potential an enemy base at (objX,objY) generates in the point (x,y).
		*/
		real8 calcEnemyBaseCharge(int x, int y, int objX, int objY, bool inside);

		/*
			Predicts the next position (x or y) a unit might move to.
		*/
		int predictNext(uint4 game_tick, int p_minusOne, int p_minusTwo, int gametick_minusOne, int gametick_minusTwo);

		Random rand;
		real8 diffScoreSum;
};

#endif
