#include "TotalPF.H"
#include "Game.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "StaticMap.H"

using namespace std;

const int TotalPF::CMTILE_WIDTH = 8;
const int TotalPF::CHARGE_IMPASSABLE_TILES = -1000;

/*
	Constructor.
*/
TotalPF::TotalPF(GameStateModulePtr mState) {
	int tile_w = mState->get_game().get_tile_points();

	width = mState->get_game().get_map().get_width() * tile_w / CMTILE_WIDTH;
	half_width = width / 2;
	height = mState->get_game().get_map().get_height() * tile_w / CMTILE_WIDTH;

	for (int cX = 0; cX < width; cX++) {
		for (int cY = 0; cY < height; cY++) {
			map[cX][cY].terrainCharge = 0;
		}
	}
}

/*
	Destructor.
*/
TotalPF::~TotalPF() {
	
}

/*
	Updates the charge a terrain tile has in a specific point.
*/
void TotalPF::updateTerrainCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	if (charge < map[x][y].terrainCharge) {
		map[x][y].terrainCharge = charge;
	}
}

/*
	Sets the charge a terrain tile has in a specific point.
	Previous charge in the point is discarded.
*/
void TotalPF::setTerrainCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].terrainCharge = charge;
}

/*
	Returns the terrain charge in a specific point.
	This method is only used in debug purpose.
*/
real8 TotalPF::getTerrainCharge(sint4 x, sint4 y) {
	if (x < 0 || x >= width - 1) return 0;
	if (y < 0 || y >= height - 1) return 0;

	return map[x][y].terrainCharge;
}

/*
	Returns the potential in the point (objX, objY).
	This method does not use the potential grid in TotalPF. Instead it calculates the correct potential
	at runtime.
*/
real8 TotalPF::getP(sint4 objX, sint4 objY, bool offensive, uint4 unitId, const Game& game, PF* pf) {
	real8 p = getTerrainCharge(posToTile(objX), posToTile(objY));
	if (p > TotalPF::CHARGE_IMPASSABLE_TILES) {
		 p = p * TBConfig::params.Terrain_w;
	}
	else {
		return p;
	}

	p += pf->getSheepCharge(objX, objY, game.tick());
	p += pf->getEnemyUnitCharge(objX, objY, offensive, game.tick());
	p += pf->getUnitCharge(objX, objY, unitId, game.tick());
	p += pf->getBaseCharge(objX, objY);
	p += pf->getEnemyBaseCharge(objX, objY);
	
	return p;
}

/*
	Returns the distance between two points (x1,y1) and (x2,y2).
*/
real8 TotalPF::distance(sint4 x1, sint4 y1, sint4 x2, sint4 y2) {
	//Correct calculation. Requires lot of CPU time.
	/*real8 ASquare = ((real8)x2 - (real8)x1) * ((real8)x2 - (real8)x1);
	real8 BSquare = ((real8)y2 - (real8)y1) * ((real8)y2 - (real8)y1);
	real8 D = sqrt(ASquare + BSquare);
	return D;*/

	//Optimized version. Requires less CPU time.
	real8 lX = (real8)abs(x2 - x1);
	real8 lY = (real8)abs(y2 - y1);
	real8 max = lX;
	real8 min = lY;
	if (lY > max) {
		max = lY;
		min = lX;
	}

	int t = (int)(min / max * (real8)100);
	real8 appD = max * StaticMap::preCalc[t];
	return appD;
}

/*
	Returns the Manhatten distance (abs(x2-x1)+abs(y2-y1)) between two points (x1,y1) and (x2,y2).
*/
real8 TotalPF::MHdistance(sint4 x1, sint4 y1, sint4 x2, sint4 y2) {
	return abs(x1 - x2) + abs(y1 - y2);
}

/*
	Returns the map width in points.
*/
int TotalPF::getRealWidth() {
	return width * CMTILE_WIDTH;
}

/*
	Returns the map height in points.
*/
int TotalPF::getRealHeight() {
	return height * CMTILE_WIDTH;
}

/*
	Returns a point (x or y) to a tile by rounding down.
*/
int TotalPF::posToTile(int pos) {
	int tile = pos / TotalPF::CMTILE_WIDTH;
	return tile;
}

/*
	Returns a point (x or y) to a tile by rounding to nearest int.
*/
int TotalPF::posToTileR(int pos) {
	int tile = (int)round((real8)pos / (real8)TotalPF::CMTILE_WIDTH);
	return tile;
}
