#include "UnitAgent.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "FireCoordinator.H"
#include "Options.H"

const int UnitAgent::TANK_FIRERANGE = 112;
const int UnitAgent::TANK_HP = 150;
const int UnitAgent::TANK_ARMOR = 1;
const sint4 UnitAgent::TANK_MIN_DMG = 26;
const sint4 UnitAgent::TANK_MAX_DMG = 34;

const int UnitAgent::MARINE_FIRERANGE = 64;
const int UnitAgent::MARINE_HP = 80;
const int UnitAgent::MARINE_ARMOR = 1;
const sint4 UnitAgent::MARINE_MIN_DMG = 5;
const sint4 UnitAgent::MARINE_MAX_DMG = 7;

const int UnitAgent::WORKER_HP = 120;
const int UnitAgent::WORKER_ARMOR = 0;

const int UnitAgent::CONTROLCENTER_HP = 1200;
const int UnitAgent::CONTROLCENTER_ARMOR = 0;

const int UnitAgent::BARRACKS_HP = 1150;
const int UnitAgent::BARRACKS_ARMOR = 0;

const int UnitAgent::FACTORY_HP = 1400;
const int UnitAgent::FACTORY_ARMOR = 0;

void UnitAgent::compute_actions(GameObjPtr unit, GameStateModulePtr state, PFUpdaterPtr pfupdater, FireCoordinatorPtr coordinator, Random* rand) {
	Game& game = state->get_game();
	
	uint4 objId = game.get_cplayer_info().get_id(unit);

	//Check if this agent shall do something this frame
	real8 cProb = pfupdater->getPF()->getProbability(objId);
	real8 rnd = (real8)rand->ranged_uint4(0, 100) / (real8)100;
	//cout << "cProb=" << cProb << " rnd=" << rnd;
	if (rnd > cProb) {
		//Do nothing
		//cout << " NoAction" << endl;
		return;
	}
	//cout << " Action" << endl;

	int unitX = *unit->sod.x;
	int unitY = *unit->sod.y;
	
	//Get which mode (offensive/defensive) to use
	bool offensive = getMode(unit, state);
	
	//Fire if any targets are within range
	bool fired = canFire(unit, unitX, unitY, objId, state, offensive, coordinator);
	if (fired) {
		return;
	}
	
	//No targets to fire at, check if we can move
	bool moved = canMove(unit, unitX, unitY, objId, offensive, pfupdater, game);
}

bool UnitAgent::canDoMove(uint4 objId, int fromX, int fromY, int toX, int toY, int unitType, PFUpdaterPtr pfupdater, const Game& game) {
	//Check boundaries
	if (toX < 4) return false;
	if (toX > pfupdater->getTotalPF()->getRealWidth() - 5) return false;
	if (toY < 4) return false;
	if (toY > pfupdater->getTotalPF()->getRealHeight() - 5) return false;

	int diffX = toX - fromX;
	int diffY = toY - fromY;
	int divider = max(abs(diffX), abs(diffY));

	int cnt = 1; //Don't check the units current position (it is passable since the unit is there...)
	int cX = fromX;
	int cY = fromY;

	real8 limit = (real8)TotalPF::CHARGE_IMPASSABLE_TILES * 0.5;
	
	while (cnt <= divider) {
		int mX = cX + calcNext(cnt, diffX, divider);
		int mY = cY + calcNext(cnt, diffY, divider);

		real8 charge = pfupdater->getTotalPF()->getTerrainCharge(mX, mY, unitType, false, game);
		//real8 charge = pfupdater->getTotalPF()->getUnitP(mX, mY, true, objId, Blackboard::TYPE_TANK, game, pfupdater->getPF());
		if (charge <= limit) {
			//Impassable tile found
			
			//Special case: If units are at the edges of the map, it can always move away from the edge.
			//This avoids units getting stuck at edges
			if (boundaryMove(fromX, fromY, toX, toY, pfupdater->getTotalPF())) {
				return true;
			}

			return false;
		}
		
		cnt++;
	}

	return true;
}

bool UnitAgent::boundaryMove(int fromX, int fromY, int toX, int toY, TotalPFPtr totalpf) {
	int d = 10;

	if (fromY == toY) {
		//Check vertical boundary move
		if (toX > fromX && toX <= d) {
			return true;
		}
		if (toX < fromX && toX >= totalpf->getRealWidth() - d) {
			return true;
		}
	}

	if (fromX == toX) {
		//Check horisontal boundary move
		if (toY > fromY && toY <= d) {
			return true;
		}
		if (toY < fromY && toY >= totalpf->getRealHeight() - d) {
			return true;
		}
	}

	return false;
}

int UnitAgent::calcNext(int cnt, int diffVal, int divider) {
	real8 next = (real8) cnt * (real8)diffVal / (real8)divider;

	int val = (int)(next * 10);
	int res = (int)next;

	if (val % 10 >= 5) res++;

	return res;
}

bool UnitAgent::canMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, PFUpdaterPtr pfupdater, const Game& game) {
	//If no targets are within range, check if there are better positions to move to
	TotalPFPtr pf = pfupdater->getTotalPF();

	int maxSpeed = *unit->sod.max_speed;
	
	//Calculate which tiles to check
	int x1 = unitX - maxSpeed;
	int x2 = unitX + maxSpeed;
	int y1 = unitY - maxSpeed;
	int y2 = unitY + maxSpeed;

	//Iterate through the tiles to get the highest (=best) charge
	//Start with the charge for current tile
	real8 bestCharge = pf->getUnitP(unitX, unitY, offensive, objId, Blackboard::getType(unit), game, pfupdater->getPF());
	
	int bestX = unitX;
	int bestY = unitY;
	
	//Get the surrounding tile with highest charge
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 penalty = 1;
#if USE_CLOSEST_BASE_POINT == 1
			if (cX != unitX && cY != unitY) {
				//Give a small penalty for moving diagonally
				penalty = 0.9999;
			}
#endif
			if (cX != unitX || cY != unitY) {
				if (canDoMove(objId, unitX, unitY, cX, cY, Blackboard::getType(unit), pfupdater, game)) {
					real8 cCharge = pf->getUnitP(cX, cY, offensive, objId, Blackboard::getType(unit), game, pfupdater->getPF()) * penalty;
					if (cCharge > bestCharge) {
						bestCharge = cCharge;
						bestX = cX;
						bestY = cY;
					}
				}
			}
		}
	}

	//If we've found a surrounding tile with higher charge than current tile, move there
	if (bestX != unitX || bestY != unitY) {
		int useSpeed = *unit->sod.max_speed;
		if (pf->distance(unitX, unitY, bestX, bestY) < useSpeed) {
			useSpeed = pf->distance(unitX, unitY, bestX, bestY);
		}
		pfupdater->getPF()->setNextUnitPosition(objId, bestX, bestY);

		Vector<sint4> params;
		params.push_back(bestX);
		params.push_back(bestY);
		params.push_back(useSpeed);

		unit->set_action("move", params);

		//Unit has moved
		return true;
	}

	//Unit has not moved
	return false;
}

/*
	Decides which mode (offensive = true, defensive = false) to use for a tank.
*/
bool UnitAgent::getMode(GameObjPtr unit, GameStateModulePtr state) {
	bool offensive = true;

	if (TBConfig::hasAddition3()) {
		Game& game = state->get_game();
		//Addition #4: Units are always in defensive move (tries to stay out of reach for fire) if in cooldown
		sint4 lastFired = unit->component("weapon")->get_int("last_fired");
		if (lastFired >= 0) {
			uint4 game_tick = game.tick();
			sint4 cooldown = unit->component("weapon")->get_int("cooldown");
			int timeToShoot = cooldown - (game_tick - (uint4)lastFired);
			if (timeToShoot > 0) {
				offensive = false;
				return offensive;
			}
		}
	}

	return offensive;
}

bool UnitAgent::canFire(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, FireCoordinatorPtr coordinator) {
	Game& game = state->get_game();
	
	//First, check time to shoot
	sint4 lastFired = unit->component("weapon")->get_int("last_fired");
	if (lastFired >= 0) {
		uint4 game_tick = game.tick();
		sint4 cooldown = unit->component("weapon")->get_int("cooldown");
		int timeToShoot = cooldown - (game_tick - (uint4)lastFired);

		if (timeToShoot > 0) {
			return false;
		}
	}

	std::string unitType = unit->bp_name();

	//Get the target for this unit from the global coordinator
	//-1 = No target available.
	int targetId = coordinator->getTargetId(objId);
	GameObjPtr target = coordinator->getTargetObj(objId);
	if (targetId >= 0) {

		std::string targType = target->bp_name();

		//cout << objId << "[" << unitType << "] --> " << targetId << "[" << targType << "]" << endl;

		//ATTACK!
		Vector<sint4> args;
		args.push_back(targetId);
		unit->component("weapon")->set_action("attack", args);
		return true;
	}

	return false;
}

